"""Header generator for ADAU14xx parts."""
from .header import Field, PacketHeader, PacketHeaderGenerator


class Adau14xxHeaderGenerator(PacketHeaderGenerator):
    """Header generator for ADAU14xx parts."""

    @staticmethod
    def new_write_header() -> PacketHeader:
        """Generate a new header for a ADAU14xx write packet."""
        return PacketHeader(
            [
                Field("operation", 0, 1),
                Field("safeload", 1, 1),
                Field("channel", 2, 1),
                Field("total_length", 3, 4),
                Field("chip_address", 7, 1),
                Field("data_length", 8, 4),
                Field("address", 12, 2),
            ]
        )

    @staticmethod
    def new_read_request_header() -> PacketHeader:
        """Generate a new header for a ADAU14xx read request packet."""
        return PacketHeader(
            [
                Field("operation", 0, 1),
                Field("total_length", 1, 4),
                Field("chip_address", 5, 1),
                Field("data_length", 6, 4),
                Field("address", 10, 2),
                Field("reserved", 12, 2),
            ]
        )

    @staticmethod
    def new_read_response_header() -> PacketHeader:
        """Generate a new header for a ADAU14xx read response packet."""
        return PacketHeader(
            [
                Field("operation", 0, 1),
                Field("total_length", 1, 4),
                Field("chip_address", 5, 1),
                Field("data_length", 6, 4),
                Field("address", 10, 2),
                Field("success", 12, 1),
                Field("reserved", 13, 1),
            ]
        )



"""Factory module for generating DSP objects."""
import logging
from typing import Dict, Type, Union

from sigmadsp.protocols.i2c import I2cProtocol
from sigmadsp.protocols.spi import SpiProtocol

from .adau1x01 import Adau1x01
from .adau14xx import Adau14xx
from .common import ConfigurationError, Dsp, InputPin, OutputPin

logger = logging.getLogger(__name__)

VALID_ADAU14XX = ["adau14xx", "adau145x", "adau146x", "adau147x"]
VALID_ADAU1X01 = ["adau1x01", "adau1401", "adau1701"]


def dsp_factory(dsp_type_name: str) -> Type[Dsp]:
    """Creates Dsp objects, based on the name of the DSP.

    Args:
        dsp_type_name (str): The name of the DSP.

    Returns:
        Type[Dsp]: The matching DSP class.
    """
    if dsp_type_name in VALID_ADAU14XX:
        return Adau14xx

    elif dsp_type_name in VALID_ADAU1X01:
        return Adau1x01

    else:
        raise TypeError("DSP type {dsp_type} is not known.")


def dsp_from_config(config: Dict) -> Dsp:
    """Parse a configuration dictionary and create a new DSP from it.

    Args:
        config (dict): The configuration dictionary.

    Returns:
        Dsp: The created DSP object.
    """
    dsp: Dsp

    try:
        dsp_protocol_name: str = config["dsp"]["protocol"].lower()
        dsp_type_name: str = config["dsp"]["type"].lower()
        bus = int(config["dsp"]["bus_number"])
        device = int(config["dsp"]["device_address"])

    except KeyError as e:
        logger.error("Key %s missing from the DSP configuration.", e.args[0])
        raise ConfigurationError from e

    else:
        # Generate the protocol first (e.g. SPI or I2C).
        dsp_protocol: Union[SpiProtocol, I2cProtocol]

        if dsp_protocol_name == "spi":
            dsp_protocol = SpiProtocol(bus=bus, device=device)

        elif dsp_protocol_name == "i2c":
            dsp_protocol = I2cProtocol(bus=bus, device=device)

        else:
            raise TypeError(f"Unknown DSP protocol {dsp_protocol_name}.")

        # Then, generate the dsp itself.
        dsp_type: Type[Dsp] = dsp_factory(dsp_type_name)

        # Use safeload by default.
        dsp = dsp_type(True, dsp_protocol)

    try:
        # Parse the configuration for pin definitions.
        for pin_definition_key in config["dsp"]["pins"]:
            pin_definition = config["dsp"]["pins"][pin_definition_key]

            if pin_definition["mode"] == "output":
                output_pin = OutputPin(
                    pin_definition_key,
                    pin_definition["number"],
                    pin_definition["initial_state"],
                    pin_definition["active_high"],
                )

                dsp.add_pin(output_pin)

            elif pin_definition["mode"] == "input":
                input_pin = InputPin(
                    pin_definition_key,
                    pin_definition["number"],
                    pin_definition["pull_up"],
                    pin_definition["active_state"],
                    pin_definition["bounce_time"],
                )

                dsp.add_pin(input_pin)

    except (KeyError, TypeError):
        logger.info("No DSP pin definitions were found in the configuration file.")

    return dsp



"""This module implements the base class for I2C and SPI communication handlers."""
import logging
import threading
from abc import ABC, abstractmethod
from multiprocessing import Queue

from sigmadsp.sigmastudio.common import ReadRequest, WriteRequest

# A logger for this module
logger = logging.getLogger(__name__)


class DspProtocol(ABC):
    """Base class for communication handlers talking to SigmaDSP chipsets."""

    def run(self):
        """Start the DSP protocol thread."""
        # Generate queues, for communicating with the protocol handler thread within this class.
        self.transmit_queue = Queue()
        self.receive_queue = Queue()

        protocol = self.__class__.__name__

        logger.info("Starting %s handling thread.", protocol)
        self.thread = threading.Thread(target=self.serve_forever, name=f"{protocol} handler thread", daemon=True)
        self.thread.start()

    def write(self, address: int, data: bytes):
        """Write data over the hardware interface via the pipe to the hardware thread.

        Args:
            address (int): DSP register address to write to
            data (bytes): Binary data to write
        """
        self.transmit_queue.put(WriteRequest(address, data))

    def read(self, address: int, length: int) -> bytes:
        """Read data from the hardware interface via the pipe to the hardware thread.

        Args:
            address (int): DSP register address to read from
            length (int): Number of bytes to read

        Returns:
            bytes: Register content
        """
        self.transmit_queue.put(ReadRequest(address, length))
        return self.receive_queue.get()

    def serve_forever(self):
        """Handle incoming requests for writing or reading data."""
        while True:
            try:
                request = self.transmit_queue.get()

            except EOFError:
                break

            if isinstance(request, WriteRequest):
                self._write(request.address, request.data)

            elif isinstance(request, ReadRequest):
                data = self._read(request.address, request.length)

                try:
                    self.receive_queue.put(data)

                except EOFError:
                    break

    @abstractmethod
    def _read(self, address: int, length: int) -> bytes:
        """Read data from a SigmaDSP.

        Args:
            address (int): Address to read from
            length (int): Number of bytes to read

        Returns:
            bytes: Data that was read from the DSP
        """

    @abstractmethod
    def _write(self, address: int, data: bytes) -> int:
        """Write data onto a SigmaDSP.

        Args:
            address (int): Address to write to
            data (bytes): Data to write

        Returns:
            int: Number of written bytes.
        """
